from yfs.asset_types import AssetTypes
from yfs.exchanges import ExchangeTypes
from yfs.lookup import fuzzy_search
from yfs.options import get_options_page
from yfs.statistics import get_statistics_page, get_multiple_statistics_pages
from yfs.summary import get_summary_page, get_multiple_summary_pages

__all__ = [
    "AssetTypes",
    "ExchangeTypes",
    "fuzzy_search",
    "get_options_page",
    "get_statistics_page",
    "get_multiple_statistics_pages",
    "get_summary_page",
    "get_multiple_summary_pages",
]
__version__ = "0.3.0"



"""Send get requests."""

from typing import Dict

import requests
from requests import Response, Session


def requestor(
    url: str, session: Session = None, proxies: Dict[str, str] = None, timeout: int = 5
) -> Response:
    """Send get requests.

    Args:
        url (str): The url to send a request to.
        session (Session): A Session object to send a request with.
        proxies (dict): Dictionary mapping protocol to the URL of the proxy.
        timeout (int): How long to wait for the server to send a response.
    """
    # TODO: try and pass a session with retries plus whaor. pylint: disable=W0511
    if session:
        return session.get(url, proxies=proxies, timeout=timeout)

    return requests.get(url, proxies=proxies, timeout=timeout)



"""A module for cleaning tables, fields and values."""

from functools import partial
from typing import Dict, Optional, Union

import pendulum
from pendulum import DateTime
from pydantic import validator
from requests_html import HTML


numbers_with_suffix = {
    "T": 1_000_000_000_000,
    "B": 1_000_000_000,
    "M": 1_000_000,
    "K": 1_000,
}


def field_cleaner(field: str) -> str:
    """Convert field string from an html response into a snake case variable.

    Args:
        field (str): A dirty field string.

    Example:
        |Input             |  Output                |
        |------------------|------------------------|
        |Previous Close    | previous_close         |
        |Avg. Volume       | avg_volume             |
        |Beta (5Y Monthly) | beta_five_year_monthly |
        |PE Ratio (TTM)    | pe_ratio_ttm           |

    Returns:
        str: lowercased and converted to snake case.
    """
    return (
        field.lower()
        .replace("(", "")
        .replace(")", "")
        .replace("'", "")
        .replace(".", "")
        .replace("& ", "")
        .replace("-", "")
        .replace("52", "fifty_two")
        .replace("5y", "five_year")
        .replace("1y", "one_year")
        .replace(" ", "_")
        .replace("5_yr", "five_year")
        .replace("5_year", "five_year")
        .replace("200day", "two_hundred_day")
        .replace("50day", "fifty_day")
        .replace("/", "_")
        .replace("&", "")
        .replace("%", "percent")
        .replace("10_day", "ten_day")
        .replace("3_month", "three_month")
        .strip()
    )


def table_cleaner(html_table: HTML) -> Optional[Dict]:
    """Clean table with two fields.

    Args:
        html_table (HTML): HTML object parsed from a table section.

    Returns:
        dict: cleaned fields (keys) and string (values).
        None: if html_table does not contain table elements.
    """
    rows = [row.text.split("\n") for row in html_table.find("tr")]
    rows = list(filter(lambda row: len(row) == 2, rows))

    data = {}

    for field, value in rows:
        field = field_cleaner(field)
        data[field] = value

    if data:
        return data

    return None


cleaner = partial(validator, pre=True, allow_reuse=True)
"""* Partial overloading the pytdanitc.validator function with common args."""


class CommonCleaners:
    """Contains the most commonly used methods for cleaning values."""

    @staticmethod
    def remove_comma(value: str) -> str:
        """Remove commas from strings and strip whitespace.

        Args:
            value (str): A number with more than 3 digits.

        Example:
            |Input   | Output|
            |--------|-------|
            |'5,000' | '5000'|

        Returns:
            str: Commas removed and whitespace stripped.
        """
        return value.replace(",", "").strip()

    @staticmethod
    def remove_brakets(value: str) -> str:
        """Remove () brakets from strings and strip whitespace.

        Args:
            value (str): Containing () brakets. Normally surrounding a percent change.

        Example:
            |Input           |Output       |
            |----------------|-------------|
            |+19.60 (+1.36%) |+19.60 +1.36 |

        Returns:
            str: () brakets removed and whitespace stripped.
        """
        return value.replace("(", "").replace(")", "").strip()

    @staticmethod
    def remove_percent_sign(value: str) -> str:
        """Remove percent sign % from string and strip whitespace.

        Args:
            value (str): Containing percent sign.

        Example:
            |Input     |Output |
            |----------|-------|
            |+1.36%    |+1.36  |

        Returns:
            str: Percent sign % removed and whitespace stripped.
        """
        return value.replace("%", "").strip()

    @classmethod
    def remove_brakets_and_percent_sign(cls, value: str) -> str:
        """Remove () brakets and % percent signs from string.

        Args:
            value (str): Contains () backets and % percent sign.

        Example:
            |Input              |Output      |
            |-------------------|------------|
            |+19.60 (+1.36%)    |+19.60 +1.36|

        Returns:
            str: () brakets % percent sign removed.
        """
        value = cls.remove_brakets(value)
        return cls.remove_percent_sign(value)

    @staticmethod
    def value_is_missing(value: str) -> bool:
        """Check if value has missing data.

        It checks to see if "N/A" and other missing data strings are present in the value.

        Args:
            value (str): A value parsed from yahoo finance.

        Example:
            |Input                    |Output      |
            |-------------------------|------------|
            |"Ex-Dividend Date    N/A"|True        |

        Returns:
            bool: True if a missing data string is found in value else False.

        """
        if not isinstance(value, str):
            return False

        missing_values = [
            "N/A",
            "undefined",
            "+-",
            "-+",
        ]

        for missing in missing_values:
            if missing in value:
                return True

        other_missing = ["", " ", "-"]

        for missing in other_missing:
            if value == missing:
                return True

        return False

    @staticmethod
    def has_large_number_suffix(value: str) -> bool:
        """Check if string representation of a number has T,B,M,K suffix.

        Args:
            value (str): A value to check if contains a T,B,M,K suffix.

        Example:

            |Input  |Output      |
            |-------|------------|
            |225.0M |True        |

        Returns:
            bool: True if string value contains T,B,M,K suffix else false.
        """
        for suffix, _ in numbers_with_suffix.items():
            if suffix in value.upper():
                return True

        return False

    @staticmethod
    def clean_large_number(value: str) -> Optional[int]:
        """Convert a string representation of a number with a T,B,M,K suffix to an int.

        Args:
            value (str): A value which contains T,B,M,K suffix.

        Example:
            |Input |Output       |
            |------|-------------|
            |2.5B  |2_500_000_000|

        Returns:
            int: suffix removed and used as multiplier to convert to an int.
            None: This can return None, but should not because the
                has_large_number_suffix method should be used to check if
                this method should even be run.
        """
        value = value.upper()

        for suffix, multiplier in numbers_with_suffix.items():
            if suffix in value:
                return round(float(value.replace(suffix, "")) * multiplier)

        return None

    @classmethod
    def common_value_cleaner(cls, value: str) -> Union[int, str]:
        """Most common method for cleaning most values from yahoo finance.

        Removes commas and converts number if it has a suffix.

        Args:
            value (str): value to be cleaned.

        Example:
            |Input   |Output       |Type|
            |--------|-------------|----|
            |"5,000" | "5000"      |str |
            |"2.5M"  | 2_500_000   |int |

        Returns:
            str: If value is cleaned and doesn't have a suffix.
            int: If value is cleaned and has a suffix.

        """
        value = cls.remove_comma(value)

        if cls.has_large_number_suffix(value):
            value = cls.clean_large_number(value)

        return value

    @classmethod
    def clean_common_values(cls, value: str) -> Optional[Union[int, str]]:
        """Most common vanilla method for cleaning most values with a check for missing values.

        Args:
            value (str): value to be cleaned.

        Example:
            |Input      |Output    |Type|
            |-----------|----------|----|
            |"5,000"    |"5000"    |str |
            |"2.5M"     |2_500_000 |int |
            |"N/A"      |None      |None|

        Returns:
            str:  If value is cleaned and doesn't have a suffix.
            int:  If value is cleaned and has a suffix.
            None: If value is missing.
        """
        if cls.value_is_missing(value):
            return None
        return cls.common_value_cleaner(value)

    @classmethod
    def clean_basic_percentage(cls, value: str) -> Optional[str]:
        """Clean a single percentage value.

        Args:
            value (str): value to be cleaned.

        Example:
            |Input   |Output |Type|
            |--------|-------|----|
            |"-3.4%" |"-3.4" |str |
            |"N/A"   |None   |None|

        Returns:
            str: cleaned value if value is not missing.
            None: if value is missing.
        """
        if cls.value_is_missing(value):
            return None
        value = cls.remove_percent_sign(value)
        return cls.remove_comma(value)

    @classmethod
    def clean_date(cls, value: str) -> DateTime:
        """Clean and convert a string date.

        Uses pendulum parse method to extract datetime information. Sometimes yahoo
        finance give multiple dates in one value field. This normally happens in the
        Earnings Date section on the Summary page. The Earnings Date may have a single
        date or an estimated Earnings Date range. It would be very easy to have this method
        output a pendulum.period.Period object that can represent the date range. The
        decision was made to just return the start of the earnings period for consistency.

        Args:
            value (str): Date string to be converted to datetime object.

        Example:
            |Input                                         |Output                            |
            |----------------------------------------------|----------------------------------|
            |"Earnings Date    Oct 26, 2020 - Oct 30, 2020"|DateTime 2020-10-26T00:00:00+00:00|

        Returns:
            str: cleaned if value is not missing.
            None: if value is missing.
        """
        if cls.value_is_missing(value):
            return None

        dates = value.split("-")

        if len(dates) > 1:
            start, _ = dates
            start = pendulum.parse(start, strict=False)
            return start

        return pendulum.parse(value, strict=False)

    @classmethod
    def clean_symbol(cls, value: str) -> str:
        """Make symbol uppercase.

        Args:
            value (str): Stock symbol.

        Example:
            |Input |Output |
            |------|-------|
            |aapl  |AAPL   |

        Returns:
            str: Uppercased.
        """
        return value.upper()

    @classmethod
    def clean_first_value_split_by_dash(cls, value: str) -> str:
        """Split value separated by '-' and return the first value.

        Args:
            value (str): A string with multiple values separated by a '-'.

        Example:
            |Input            |Output  |
            |-----------------|--------|
            |"2.3400 - 2.4900"|"2.3400"|

        Returns:
            strs: First value parse from string.
            None: if value is missing.
        """
        if cls.value_is_missing(value):
            return None

        value, _ = value.split("-")
        return cls.common_value_cleaner(value)

    @classmethod
    def clean_second_value_split_by_dash(cls, value: str) -> str:
        """Split value separated by '-' and return the second value.

        Args:
            value (str): A string with multiple values separated by a '-'.

        Example:
            |Input            |Output  |
            |-----------------|--------|
            |"2.3400 - 2.4900"|"2.4900"|

        Returns:
            strs: Second value parse from string.
            None: if value is missing.
        """
        if cls.value_is_missing(value):
            return None

        _, value = value.split("-")
        value = cls.common_value_cleaner(value)
        return value

    @classmethod
    def clean_first_value_split_by_space(cls, value: str) -> str:
        """Clean first string containing a change and percent.

        This will strip brakets and percent sign leaving only a whitespace
        between the change and percent change. Next it will split the values then
        return the first value. Normally the Forward Dividend & Yield from the
        summary page.

        Args:
            value (str): Normally a string containing change and percent change.

        Example:
            |Input            |Output  |
            |-----------------|--------|
            |"0.82 (0.73%)"   |"0.82"  |

        Returns:
            str: The first string.
            None: if value is missing.
        """
        if cls.value_is_missing(value):
            return None

        value, _ = cls.remove_brakets_and_percent_sign(value).split(" ")
        return value

    @classmethod
    def clean_second_value_split_by_space(cls, value: str) -> str:
        """Clean second string containing a change and percent.

        This will strip brakets and percent sign leaving only a whitespace
        between the change and percent change. Next it will split the values then
        return the second value. Normally the Forward Dividend & Yield from the
        summary page.

        Args:
            value (str): Normally a string containing change and percent change.

        Example:
            |Input            |Output  |
            |-----------------|--------|
            |"0.82 (0.73%)"   |"0.73"  |

        Returns:
            str: The Second value.
            None: if value is missing.
        """
        if cls.value_is_missing(value):
            return None

        _, percentage = cls.remove_brakets_and_percent_sign(value).split(" ")
        return percentage

    @classmethod
    def clean_first_value_split_by_x(cls, value: str) -> str:
        """Split value separated by 'x' and return the first value.

        Args:
            value (str): A string with multiple values separated by a 'x'.

        Example:
            |Input            |Output  |
            |-----------------|--------|
            |"2.4000 x 21500" |"2.4000"|

        Returns:
            strs: First value parse from string.
            None: if value is missing.
        """
        if cls.value_is_missing(value):
            return None

        price, _ = value.split("x")
        return cls.common_value_cleaner(price)

    @classmethod
    def clean_second_value_split_by_x(cls, value: str) -> str:
        """Split value separated by 'x' and return the second value.

        Args:
            value (str): A string with multiple values separated by a 'x'.

        Example:
            |Input            |Output  |
            |-----------------|--------|
            |"2.4000 x 21500" |"21500" |

        Returns:
            strs: Second value parse from string.
            None: if value is missing.
        """
        if cls.value_is_missing(value):
            return None

        _, volume = value.split("x")
        return cls.common_value_cleaner(volume)



"""Contains the classes and functions for scraping a yahoo finance option page."""

from enum import Enum
from itertools import cycle
from typing import Iterable, List, Optional, Tuple, Union

import pandas
from pandas import DataFrame
import pendulum
from pendulum.datetime import DateTime
from pydantic import BaseModel as Base
from pydantic import Field, validator
from requests_html import HTML

from .cleaner import cleaner, CommonCleaners, field_cleaner
from .lookup import fuzzy_search
from .requestor import requestor


class ContractExpiration(Base):
    """Contract Expiration.

    Attributes:
        symbol (str): Ticker symbol.
        timestamp (str): Timestamp of expiration date.
        expiration_date (DateTime): Datetime of expiration date.

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.
    """

    symbol: str
    timestamp: str
    expiration_date: DateTime = Field(alias="timestamp")

    @validator("expiration_date", pre=True)
    def convert_to_datetime(cls, value: str) -> DateTime:  # pylint: disable=E0213,R0201
        """Convert expiration timestamp to datetime."""
        expiration_date = pendulum.from_timestamp(int(value), tz="UTC")
        return expiration_date

    def __lt__(self, other: "ContractExpiration") -> Optional["ContractExpiration"]:
        """Compare expiration_dates for sorting."""
        if other.__class__ is self.__class__:
            return self.expiration_date < other.expiration_date
        return None


class ContractExpirationList(Base):
    """Contains Multiple Expirations.

    Attributes:
        expiration_list (List[ContractExpiration]): multiple expirations.

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.

    """

    expiration_list: List[ContractExpiration]

    @validator("expiration_list")
    def sort_dates(  # pylint: disable=E0213,R0201
        cls, values: List[ContractExpiration]
    ) -> List[ContractExpiration]:
        """Sort expiration_list by date."""
        return sorted(values)

    def filter_expirations_after(self, after: DateTime) -> None:
        """Filter out any expiration dates prior to the after date.

        Args:
            after (DateTime): datetime to filter.

        Example:
            |Input                      | Args         |Output            |
            |---------------------------|--------------|------------------|
            |[01JAN19, 01FEB19, 01MAR19]|after: 15JAN19|[01FEB19, 01MAR19]|
        """
        filtered = list(filter(lambda exp: exp.expiration_date >= after, self.expiration_list))
        self.expiration_list = filtered

    def filter_expirations_before(self, before: DateTime) -> None:
        """Filter out any expiration dates post the before date.

        Args:
            before (DateTime): datetime to filter.

        Example:
            |Input                      | Args          |Output   |
            |---------------------------|---------------|---------|
            |[01JAN19, 01FEB19, 01MAR19]|before: 15JAN19|[01JAN19]|

        """
        filtered = list(filter(lambda exp: exp.expiration_date <= before, self.expiration_list))
        self.expiration_list = filtered

    def filter_expirations_between(self, after: DateTime, before: DateTime) -> None:
        """Filter dates outside of a after and before range.

        Args:
            after (DateTime): datetime to filter.
            before (DateTime): datetime to filter.

        Example:
            |Input                      | Args                         |Output            |
            |---------------------------|------------------------------|------------------|
            |[01JAN19, 01FEB19, 01MAR19]|after: 15JAN19,before: 15JAN19|[01FEB19, 01MAR19]|
        """
        self.filter_expirations_after(after=after)
        self.filter_expirations_before(before=before)

    def filter_expirations_after_days(self, days: int) -> None:
        """Filter expirations only allowing expirations after n days.

        Args:
            days (int): Number of days to start filtering from. All expirations
                which expire prior to the days will be filtered out.
        """
        after = pendulum.now().add(days=days)
        self.filter_expirations_after(after=after)

    def filter_expirations_before_days(self, days: int) -> None:
        """Filter expiration only allowing expirations before n days.

        Args:
            days (int): Number of days to start filtering from. All expirations
                which expire post days will be filtered out.
        """
        before = pendulum.now().add(days=days)
        self.filter_expirations_before(before=before)

    def filter_expirations_between_days(
        self, after_days: Optional[int] = None, before_days: Optional[int] = None
    ) -> None:
        """Filter expiration only allowing expirations between a range of days.

        Args:
            after_days (int): Number of days to start filtering from. All expirations
                which expire prior to the days will be filtered out.
            before_days (int): Number of days to start filtering from. All expirations
                which expire post days will be filtered out.
        """
        if after_days:
            self.filter_expirations_after_days(days=after_days)

        if before_days:
            self.filter_expirations_before_days(days=before_days)

    def __len__(self) -> int:
        """Length of the expiration_list."""
        return len(self.expiration_list)

    def __iter__(self) -> Iterable:
        """Iterate over the expirations_list."""
        return iter(self.expiration_list)

    def __add__(self, other: "ContractExpirationList") -> Optional["ContractExpirationList"]:
        """Combine two ContractExpirationLists using the + operator."""
        if self.__class__ == other.__class__:
            expiration_list = self.expiration_list + other.expiration_list
            return ContractExpirationList(expiration_list=expiration_list)
        return None


class OptionContractType(str, Enum):
    """Enum for option contract types."""

    CALL = "call"
    PUT = "put"


class OptionContract(Base):
    """Represents an Option Contract.

    Attributes:
        symbol (str): Ticker symbol.
        contract_type (OptionContractType): Call or Put type.

        timestamp (str): Raw timestamp scraped from yahoo finance. This string is left
            untouched to make sure there is no issues when building a URL.
        expiration_date (DateTime): Converted from the timestamp. This allows allows
            sorting and filtering.
        in_the_money (bool): True if strike price is ITM else False.

        contract_name (str): Contract Name.
        last_trade_date (DateTime): Date of last trade.
        strike (float): Contracts strike price.
        last_price (float): Last price of a transaction between a contract buyer and a seller.

        bid (float): Last bid price.
        ask (float): Last ask price.

        change (float): Price change in dollars.
        percent_change (float): Price change in percentage.
        volume (int): Volume.
        open_interest (int): Number of contracts opened.
        implied_volatility (float): Contract IV.

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.
    """

    symbol: str
    contract_type: OptionContractType
    timestamp: str
    expiration_date: DateTime  # already parsed from ContractExpiration
    in_the_money: bool

    contract_name: str
    # last_trade_date: Optional[DateTime] #TODO: fix validator
    strike: Optional[float]
    last_price: Optional[float]

    bid: Optional[float]
    ask: Optional[float]

    change: Optional[float]
    percent_change: Optional[float]
    volume: Optional[int]
    open_interest: Optional[int]
    implied_volatility: Optional[float]

    # @cleaner("last_trade_date")
    # def clean_last_trade_date(cls, value):
    #     return pendulum.parse(value, strict=False)

    _clean_common_values = cleaner(
        "strike", "last_price", "bid", "ask", "volume", "open_interest"
    )(CommonCleaners.clean_common_values)

    _clean_percentages = cleaner("change", "percent_change", "implied_volatility")(
        CommonCleaners.clean_basic_percentage
    )

    class Config:  # noqa: D106 pylint: disable=C0115
        use_enum_values = True


class OptionsChain(Base):
    """Chain of option contracts with the same expiration date.

    Attributes:
        symbol (str): Company symbol.
        expiration_date (DateTime): Contracts expiration date.
        chain (List[OptionContract]): List of OptionContracts.

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.
    """

    symbol: str
    expiration_date: DateTime
    chain: List[OptionContract]

    @property
    def dataframe(self) -> DataFrame:
        """Return a dataframe of the option chain."""
        data = self.dict()
        chain_data = data["chain"]
        dataframe = DataFrame.from_dict(chain_data)
        return dataframe

    @property
    def calls(self) -> "OptionsChain":
        """Return a OptionChain with only call contracts."""
        call_chain = list(
            filter(lambda contract: contract.contract_type == OptionContractType.CALL, self.chain)
        )
        return OptionsChain(
            symbol=self.symbol, expiration_date=self.expiration_date, chain=call_chain
        )

    @property
    def puts(self) -> "OptionsChain":
        """Return a OptionChain with only put contracts."""
        put_chain = list(
            filter(lambda contract: contract.contract_type == OptionContractType.PUT, self.chain)
        )
        return OptionsChain(
            symbol=self.symbol, expiration_date=self.expiration_date, chain=put_chain
        )

    def __len__(self) -> int:
        """Return the number of OptionContracts in the OptionChain."""
        return len(self.chain)


class MultipleOptionChains(Base):
    """Multiple Option Chains with multiple expiration dates.

    Attributes:
        option_chain_list (List[OptionsChain]): List of option chains.
        contract_expiration_list (ContractExpirationList): List of expirations.

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.
    """

    option_chain_list: List[OptionsChain]
    contract_expiration_list: ContractExpirationList

    @property
    def dataframe(self) -> DataFrame:
        """Return a dataframe of multiple option chains."""
        if len(self.option_chain_list) == 1:
            return self.option_chain_list[0].dataframe

        dataframes = []

        for option_chain in self.option_chain_list:
            dataframes.append(option_chain.dataframe)

        return pandas.concat(dataframes, ignore_index=True)

    @property
    def calls(self) -> "MultipleOptionChains":
        """Return a MultipleOptionChains object with only call contracts."""
        calls = [chain.calls for chain in self]
        return MultipleOptionChains(
            option_chain_list=calls, contract_expiration_list=self.contract_expiration_list
        )

    @property
    def puts(self) -> "MultipleOptionChains":
        """Return a MultipleOptionChains object with only put contracts."""
        puts = [chain.puts for chain in self]
        return MultipleOptionChains(
            option_chain_list=puts, contract_expiration_list=self.contract_expiration_list
        )

    def __len__(self) -> int:
        """Return the number of option chains."""
        return len(self.option_chain_list)

    def __iter__(self) -> Iterable:
        """Iterate over option chain list."""
        return iter(self.option_chain_list)

    def __add__(self, other: "MultipleOptionChains") -> Optional["MultipleOptionChains"]:
        """Concatenate MultipleOptionChains."""
        if self.__class__ == other.__class__:
            option_chain_list = self.option_chain_list + other.option_chain_list
            contract_expiration_list = (
                self.contract_expiration_list + other.contract_expiration_list
            )
            return MultipleOptionChains(
                option_chain_list=option_chain_list,
                contract_expiration_list=contract_expiration_list,
            )
        return None  # NOTE: Maybe Should Raise here


def get_table_elements(html: HTML) -> Tuple[Optional[HTML], Optional[HTML]]:
    """Parse call and put HTML table elements.

    Args:
        html (HTML): HTML element with call and put data.

    Returns:
        Tuple of found call and put html elements.
    """
    calls_table = html.find("table.calls", first=True)
    puts_table = html.find("table.puts", first=True)
    return calls_table, puts_table


def parse_option_table(
    contract_expiration: ContractExpiration, contract_type: OptionContractType, options_table: HTML
) -> List[OptionContract]:
    """Parse and clean fields and rows of a options table HTML element.

    Args:
        contract_expiration (ContractExpiration): Used to pass ContractExpiration data
            to the returned OptionContract object.
        contract_type (OptionContractType): Call or Put
        options_table (HTML): HTML element with raw options table data.

    Returns:
        A list of OptionContracts parsed from the html options_table.
    """
    head = options_table.find("thead", first=True)
    body = options_table.find("tbody", first=True)

    headers = cycle(head.text.split("\n"))

    contracts = []

    for row in body.find("tr"):
        data = contract_expiration.dict()
        data["contract_type"] = contract_type

        if "in-the-money" in row.attrs["class"]:
            data["in_the_money"] = True
        else:
            data["in_the_money"] = False

        for value in row.text.split("\n"):
            column_name = field_cleaner(next(headers))
            data[column_name] = value

        contracts.append(OptionContract(**data))

    return contracts


def get_option_expirations(
    symbol: str, **kwargs  # noqa: ANN003
) -> Optional[ContractExpirationList]:
    """Get and parse option expiration data for the selected symbol.

    Args:
        symbol (str): Ticker symbol.
        kwargs: Pass (session, proxies, and timeout) to the requestor function.

    Returns:
        ContractExpirationList
    """
    url = f"https://finance.yahoo.com/quote/{symbol}/options?p={symbol}"

    response = requestor(url, **kwargs)

    if response.ok:

        html = HTML(html=response.text, url=url)

        elements = html.find(r"div.Fl\(start\).Pend\(18px\)", first=True)

        if elements:

            timestamps = [element.attrs["value"] for element in elements.find("option")]

            expiration_list = [
                ContractExpiration(symbol=symbol, timestamp=timestamp) for timestamp in timestamps
            ]

            return ContractExpirationList(expiration_list=expiration_list)

    return None


class OptionPageNotFound(AttributeError):
    """Raised when options page data is not found."""


def get_options_page(  # pylint: disable=R0913, R0914
    symbol: str,
    after_days: int = None,
    before_days: int = None,
    first_chain: bool = False,
    use_fuzzy_search: bool = True,
    page_not_found_ok: bool = False,
    **kwargs,  # noqa: ANN003
) -> Optional[Union[OptionsChain, MultipleOptionChains]]:
    """Get options data from yahoo finance options page.

    Args:
        symbol (str): Ticker symbol.
        after_days (int): Number of days to start filtering from. All expirations
            which expire prior to the days will be filtered out.
        before_days (int): Number of days to start filtering from. All expirations
            which expire post days will be filtered out.
        first_chain (bool): If True returns first chain. Else returns all found chains
            within search range.
        use_fuzzy_search (bool): If True, does a symbol lookup validation prior
            to requesting options page data.
        page_not_found_ok (bool): If True, returns None when page is not found.
        **kwargs: Pass (session, proxies, and timeout) to the requestor function.

    Returns:
        OptionsChain: If first_chain is set to True the first found OptionsChain
            within the after_days and before_days range is returned.
            This is all option contracts from a single expiration and symbol.
        MultipleOptionChains: If first_chain is set to False all OptionsChains within
            the after_days and before_days range are returned. This can have
            multiple expirations. Even if one expiration date is found
            the MultipleOptionChains object is returned.
        None: If no contracts are found and page_not_found_ok is True.

    Raises:
        OptionPageNotFound: If page_not_found_ok is False and the Options page is not found.

    """
    if use_fuzzy_search:
        fuzzy_response = fuzzy_search(symbol, first_ticker=True, **kwargs)

        if fuzzy_response:
            symbol = fuzzy_response.symbol

    expirations_list = get_option_expirations(symbol, **kwargs)

    if expirations_list is None:
        return None

    if after_days or before_days:
        expirations_list.filter_expirations_between_days(
            after_days=after_days, before_days=before_days
        )

    mutiple_option_chains = []

    for expiration in expirations_list:
        timestamp = expiration.timestamp
        symbol = expiration.symbol

        url = f"https://finance.yahoo.com/quote/{symbol}/options?date={timestamp}&p={symbol}"

        response = requestor(url, **kwargs)

        if response.ok:

            html = HTML(html=response.text, url=url)

            calls_table, puts_table = get_table_elements(html)

            if calls_table is None or puts_table is None:
                continue

            calls = parse_option_table(expiration, "call", calls_table)
            puts = parse_option_table(expiration, "put", puts_table)

            chain = calls + puts

            option_chain = OptionsChain(
                symbol=symbol, expiration_date=expiration.expiration_date, chain=chain
            )

            if first_chain:
                return option_chain

            mutiple_option_chains.append(option_chain)

    if len(mutiple_option_chains) > 0:

        return MultipleOptionChains(
            option_chain_list=mutiple_option_chains, contract_expiration_list=expirations_list
        )

    if page_not_found_ok:
        return None

    raise OptionPageNotFound(f"{symbol} options pages is not found.")



"""Contains the classes and functions for scraping a yahoo finance summary page."""

from collections import ChainMap
from typing import Dict, Iterable, List, Optional

from pandas import DataFrame
from pendulum.date import Date
from pydantic import BaseModel as Base
from pydantic import Field
from requests_html import HTML

from .cleaner import cleaner, CommonCleaners, table_cleaner
from .lookup import fuzzy_search
from .multidownloader import _download_pages_with_threads, _download_pages_without_threads
from .quote import parse_quote_header_info, Quote
from .requestor import requestor


class SummaryPage(Base):
    """Data scraped from the yahoo finance summary page.

    Attributes:
        symbol (str): Ticker Symbol
        name (str): Ticker Name

        quote (Quote): Quote header section of the page.

        open (float): Open price.
        high (float): Days high.
        low (float): Days low.
        close (float): Days close price.

        change (float): Dollar change in price.
        percent_change (float): Percent change in price.

        previous_close (float): Previous days close price.

        bid_price (float): Bid price.
        bid_size (int): Bid size.

        ask_price (float): Ask price.
        ask_size (int): Ask size.

        fifty_two_week_high (float): High of the fifty two week range.
        fifty_two_week_low (float): Low of the fifty two week range.

        volume (int): Volume.
        average_volume (int): Average Volume.

        market_cap (int): Market capitalization.

        beta_five_year_monthly (float): Five year monthly prices benchmarked against the SPY.
        pe_ratio_ttm (float): Share Price divided by Earnings Per Share trailing twelve months.
        eps_ttm (float): Earnings per share trailing twelve months.

        earnings_date (Date): Estimated earnings report release date.

        forward_dividend_yield (float): Estimated dividend yield.
        forward_dividend_yield_percentage (float): Estimated divided yield percentage.
        exdividend_date (Date): Ex-Dividend Date.

        one_year_target_est (float): One year target estimation.

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.
    """

    symbol: str
    name: str  # from quote

    quote: Quote

    open: Optional[float]
    high: Optional[float] = Field(alias="days_range")
    low: Optional[float] = Field(alias="days_range")
    close: Optional[float]  # pre-cleaned from quote

    change: Optional[float]  # pre-cleaned from quote
    percent_change: Optional[float]  # pre-cleaned from quote

    previous_close: Optional[float]

    bid_price: Optional[float] = Field(alias="bid")
    bid_size: Optional[int] = Field(alias="bid")

    ask_price: Optional[float] = Field(alias="ask")
    ask_size: Optional[int] = Field(alias="ask")

    fifty_two_week_low: Optional[float] = Field(alias="fifty_two_week_range")
    fifty_two_week_high: Optional[float] = Field(alias="fifty_two_week_range")

    volume: Optional[int]
    average_volume: Optional[int] = Field(alias="avg_volume")

    market_cap: Optional[int]

    beta_five_year_monthly: Optional[float]
    pe_ratio_ttm: Optional[float]
    eps_ttm: Optional[float]

    earnings_date: Optional[Date]

    forward_dividend_yield: Optional[float]
    forward_dividend_yield_percentage: Optional[float] = Field(alias="forward_dividend_yield")
    exdividend_date: Optional[Date]

    one_year_target_est: Optional[float]

    _clean_symbol = cleaner("symbol")(CommonCleaners.clean_symbol)

    _clean_highs = cleaner("low", "fifty_two_week_low")(
        CommonCleaners.clean_first_value_split_by_dash
    )
    _clean_lows = cleaner("high", "fifty_two_week_high")(
        CommonCleaners.clean_second_value_split_by_dash
    )
    _clean_date = cleaner("earnings_date", "exdividend_date")(CommonCleaners.clean_date)

    _clean_common_values = cleaner(
        "open",
        "previous_close",
        "market_cap",
        "volume",
        "average_volume",
        "pe_ratio_ttm",
        "beta_five_year_monthly",
        "eps_ttm",
        "one_year_target_est",
    )(CommonCleaners.clean_common_values)

    _clean_forward_dividend_yield = cleaner("forward_dividend_yield")(
        CommonCleaners.clean_first_value_split_by_space
    )

    _clean_forward_dividend_yield_percentage = cleaner("forward_dividend_yield_percentage")(
        CommonCleaners.clean_second_value_split_by_space
    )

    _clean_bid_ask_price = cleaner("bid_price", "ask_price")(
        CommonCleaners.clean_first_value_split_by_x
    )

    _clean_bid_ask_volume = cleaner("bid_size", "ask_size")(
        CommonCleaners.clean_second_value_split_by_x
    )

    def __lt__(self, other) -> bool:  # noqa: ANN001
        """Compare SummaryPage objects to allow ordering by symbol."""
        if other.__class__ is self.__class__:
            return self.symbol < other.symbol

        return None


class SummaryPageGroup(Base):
    """Group of SummaryPage objects from multiple symbols.

    Attributes:
        pages (SummaryPage):

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.

    """

    pages: List[SummaryPage] = list()

    def append(self, page: SummaryPage) -> None:
        """Append a SummaryPage to the SummaryPageGroup.

        Args:
            page (SummaryPage): A SummaryPage object to add to the group.
        """
        if page.__class__ is SummaryPage:
            self.pages.append(page)
        else:
            raise AttributeError("Can only append SummaryPage objects.")

    @property
    def symbols(self: "SummaryPageGroup") -> List[str]:
        """List of symbols in the SummaryPageGroup."""
        return [symbol.symbol for symbol in self]

    def sort(self: "SummaryPageGroup") -> None:
        """Sort SummaryPage objects by symbol."""
        self.pages = sorted(self.pages)

    @property
    def dataframe(self: "SummaryPageGroup") -> Optional[DataFrame]:
        """Return a dataframe of multiple SummaryPage objects."""
        pages = self.dict().get("pages")

        if pages:
            dataframe = DataFrame.from_dict(pages)
            dataframe.set_index("symbol", inplace=True)
            dataframe.drop("quote", axis=1, inplace=True)
            dataframe.sort_index(inplace=True)

            return dataframe  # TODO: none or nan

        return None

    def __iter__(self: "SummaryPageGroup") -> Iterable:
        """Iterate over SummaryPage objects."""
        return iter(self.pages)

    def __len__(self: "SummaryPageGroup") -> int:
        """Length of SummaryPage objects."""
        return len(self.pages)


def parse_summary_table(html: HTML) -> Optional[Dict]:
    """Parse data from summary table HTML element."""
    quote_summary = html.find("div#quote-summary", first=True)

    if quote_summary:
        return table_cleaner(quote_summary)

    return None


def get_summary_page(
    symbol: str,
    use_fuzzy_search: bool = True,
    page_not_found_ok: bool = False,
    **kwargs,  # noqa: ANN003
) -> Optional[SummaryPage]:
    """Get summary page data.

    Args:
        symbol (str): Ticker symbol.
        use_fuzzy_search (bool): If True does a symbol lookup validation prior
            to requesting summary page data.
        page_not_found_ok (bool): If True Returns None when page is not found.
        **kwargs: Pass (session, proxies, and timeout) to the requestor function.

    Returns:
        SummaryPage: When data is found.
        None: No data is found and page_not_found_ok is True.

    Raises:
        AttributeError: When a page is not found and the page_not_found_ok arg is false.
    """
    if use_fuzzy_search:
        fuzzy_response = fuzzy_search(symbol, first_ticker=True, **kwargs)

        if fuzzy_response:
            symbol = fuzzy_response.symbol

    url = f"https://finance.yahoo.com/quote/{symbol}?p={symbol}"

    response = requestor(url, **kwargs)

    if response.ok:

        html = HTML(html=response.text, url=url)

        quote_data = parse_quote_header_info(html)
        summary_page_data = parse_summary_table(html)

        if quote_data and summary_page_data:

            data = ChainMap(quote_data.dict(), summary_page_data)
            data["symbol"] = symbol
            data["quote"] = quote_data

            return SummaryPage(**data)

    if page_not_found_ok:
        return None

    raise AttributeError(f"{symbol} summary page not found.")


def get_multiple_summary_pages(  # pylint: disable=too-many-arguments
    symbols: List[str],
    use_fuzzy_search: bool = True,
    page_not_found_ok: bool = True,
    with_threads: bool = False,
    thread_count: int = 5,
    progress_bar: bool = True,
    **kwargs,  # noqa: ANN003
) -> Optional[SummaryPageGroup]:
    """Get multiple summary pages.

    Args:
        symbols (List[str]): Ticker symbols or company names.
        use_fuzzy_search (bool): If True does a symbol lookup validation prior
            to requesting data.
        page_not_found_ok (bool): If True Returns None when page is not found.
        with_threads (bool): If True uses threading.
        thread_count (int): Number of threads to use if with_threads is set to True.
        **kwargs: Pass (session, proxies, and timeout) to the requestor function.
        progress_bar (bool): If True shows the progress bar else the progress bar
            is not shown.

    Returns:
        SummaryPageGroup: When data is found.
        None: No data is found and page_not_found_ok is True.

    Raises:
        AttributeError: When a page is not found and the page_not_found_ok arg is false.
    """
    symbols = list(set(symbols))
    group_object = SummaryPageGroup
    callable_ = get_summary_page

    if with_threads:
        return _download_pages_with_threads(
            group_object,
            callable_,
            symbols,
            use_fuzzy_search=use_fuzzy_search,
            page_not_found_ok=page_not_found_ok,
            thread_count=thread_count,
            progress_bar=progress_bar,
            **kwargs,
        )
    return _download_pages_without_threads(
        group_object,
        callable_,
        symbols,
        use_fuzzy_search=use_fuzzy_search,
        page_not_found_ok=page_not_found_ok,
        progress_bar=progress_bar,
        **kwargs,
    )



"""A module for asset types."""

from enum import Enum


class AssetTypes(str, Enum):
    """An enum representing all asset types available on yahoo finance.

    AssetTypes is mostly used to filter down choices of assets from the lookup module.
    """

    CURRENCY = "Currency"
    ETF = "ETF"
    EQUITY = "Equity"
    FUND = "Fund"
    FUTURES = "Futures"
    INDEX = "Index"
    MONEY_MARKET = "MoneyMarket"
    OPTION = "Option"
    SHITCOINS = "CRYPTOCURRENCY"


VALID_ASSET_TYPES = [asset.value for asset in AssetTypes]



"""Contains the classes and functions for using the yahoo finance look up."""

from typing import Iterable, List, Optional, Union

from decouple import config
from pydantic import BaseModel as Base
from pydantic import Field, PydanticValueError, ValidationError, validator

from .asset_types import AssetTypes, VALID_ASSET_TYPES
from .exchanges import UnitedStatesExchanges, VALID_EXCHANGE_ENUM_VALUES, VALID_EXCHANGE_UNION
from .requestor import requestor


RAISE_ERROR_ON_UNKOWN_EXCHANGE_OR_ASSET = config(
    "RAISE_ERROR_ON_UNKOWN_EXCHANGE_OR_ASSET", default=False, cast=bool
)


class ExchangeNotFoundError(PydanticValueError):
    """Raised when an exchange is not found."""

    code = "not_a_valid_exchange"
    msg_template = "Found a missing exchange: {invalid_exchange}. Please raise an issue on github."


class AssetTypeNotFoundError(PydanticValueError):
    """Raised when an assset type is not found."""

    code = "not_a_valid_asset_type"
    msg_template = "Found a missing asset: {invalid_asset_type}. Please raise an issue on github."


class ValidSymbol(Base):
    """A valid symbol response from the yahoo finance quote lookup search bar.

    Attributes:
        symbol (str): A valid ticker symbol abbreviation.
        name (str): Company name.
        exchange (st): A valid market exchange.
            The exchange str will be validated against all of the valid exchange
            enums found in the yfs.exchanges module.
        asset_type (str): A valid asset type.
            The asset type will be validated against the
            yfs.asset_types.AssetTypes enum.

    Example:
    ```python
    {
        "symbol": "AAPL",
        "name": "Apple Inc.",
        "exchange": "NASDAQ",
        "asset_type": "Equity"
    }
    ```
    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.
    """

    symbol: str
    name: str
    exchange: VALID_EXCHANGE_UNION = Field(alias="exchDisp")
    asset_type: AssetTypes = Field(alias="typeDisp")

    class Config:  # noqa: D106 pylint: disable=missing-class-docstring
        allow_population_by_field_name = True

    @validator("exchange", pre=True)
    def check_exchange(cls, value: str) -> str:  # pylint: disable=no-self-argument, no-self-use
        """Check if the exchange is in one of the valid yfs.exchanges enums.

        Raises:
            ExchangeNotFoundError: If the exchange is not one yfs.exchanges enums
                and if the RAISE_ERROR_ON_UNKOWN_EXCHANGE_OR_ASSET environmental variable
                is set to True. (Default) is False.
        """
        if RAISE_ERROR_ON_UNKOWN_EXCHANGE_OR_ASSET is True:

            if value in VALID_EXCHANGE_ENUM_VALUES:
                return value
            raise ExchangeNotFoundError(invalid_exchange=value)
        return value

    @validator("asset_type", pre=True)
    def check_type(cls, value: str) -> str:  # pylint: disable=no-self-argument, no-self-use
        """Check if the asset_type is one of the valid yfs.asset_types enum.

        Raises:
            AssetTypeNotFoundError: If the asset is not found in yfs.asset_types enum
                and if the RAISE_ERROR_ON_UNKOWN_EXCHANGE_OR_ASSET environmental variable
                is set to True. (Default) is False.

        """
        if RAISE_ERROR_ON_UNKOWN_EXCHANGE_OR_ASSET is True:

            if value in VALID_ASSET_TYPES:
                return value
            raise AssetTypeNotFoundError(invalid_asset_type=value)

        return value


class ValidSymbolList(Base):
    """A list of symbol responses from the yahoo finance quote lookup search bar.

    Attributes:
        symbols (ValidSymbol): A list of ValidSymbol's.

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.

    """

    symbols: List[ValidSymbol] = Field(alias="items")

    class Config:  # noqa: D106 pylint: disable=missing-class-docstring
        allow_population_by_field_name = True

    def __getitem__(self, index: int) -> Optional[ValidSymbol]:
        """Return ValidSymbol at index position."""
        if self.symbols:
            return self.symbols[index]
        return None

    def __len__(self) -> int:
        """Return the amount of symbols."""
        return len(self.symbols)

    def __iter__(self) -> Iterable:
        """Iterate over ValidSymbol objects."""
        return iter(self.symbols)

    # IDEA: Maybe add inplace=True
    def filter_symbols(
        self, exchange_type: VALID_EXCHANGE_UNION, asset_type: AssetTypes
    ) -> "ValidSymbolList":
        """Filter symbols by exchange and asset type.

        Args:
            exchange_type: A valid exchange enum. Example: yfs.exchanges.UnitedStatesExchanges
            asset_type: A valid asset type. Example: yfs.asset_types.AssetTypes.EQUITY

        Returns:
            ValidSymbolList: A new ValidSymbolList with symbols within the exchange and
                asset filters.

        """
        symbols = filter(lambda symbol: symbol.exchange in exchange_type, self.symbols)
        symbols = list(filter(lambda symbol: symbol.asset_type == asset_type, symbols))
        return ValidSymbolList(symbols=symbols)


def fuzzy_search(
    quote_lookup: str,
    exchange_type: VALID_EXCHANGE_UNION = UnitedStatesExchanges,
    asset_type: AssetTypes = AssetTypes.EQUITY,
    first_ticker: bool = True,
    use_filter: bool = False,
    **kwargs,  # noqa: ANN003
) -> Optional[Union[ValidSymbol, ValidSymbolList]]:
    """Lookup and validate symbols or company names.

    Examples:
    ```python
    results = fuzzy_search("apple", first_ticker=False)
    print(results.json(indent=4))
    ```

    ```json
    {
        "symbols": [
            {
                "symbol": "AAPL",
                "name": "Apple Inc.",
                "exchange": "NASDAQ",
                "asset_type": "Equity"
            },
            {
                "symbol": "APLE",
                "name": "Apple Hospitality REIT, Inc.",
                "exchange": "NYSE",
                "asset_type": "Equity"
            },
            ...
        ]
    }
    ```

    ```python
    result = fuzzy_search("aapl", first_ticker=True)
    print(result.json(indent=4))
    ```

    ```json
    {
        "symbol": "AAPL",
        "name": "Apple Inc.",
        "exchange": "NASDAQ",
        "asset_type": "Equity"
    }
    ```

    Args:
        quote_lookup: The company name or symbol to search for.
        exchange_type: One of the yfs.exchanges enums. Default is UnitedStatesExchanges.
        asset_type: One of the yfs.asset_type.AssetTypes. Default is AssetTypes.EQUITY
        first_ticker: If set to true returns the first ValidSymbol in the ValidSymbolList.
            This is normally the best recommended match from the yahoo finance quote lookup.
        **kwargs: Pass (session, proxies, and timeout) to the requestor function.

    Returns:
        ValidSymbol: If first_ticker is set to true and a valid ticker response is found.
        ValidSymbolList: If first_ticker is set to false and a list of valid ticker
            responses are found.
        None: If the yahoo finance quote lookup search returns an empty response or if the
            response did not meet the exchange and asset type filtering requirements.

    Raises:
        ExchangeNotFoundError: If a exchange is found which has not been implemented in one of the
            yfs.exchanges enums. If this error is raised please raise an issue on github with the
            output.
        AssetTypeNotFoundError: If an asset type is not found in the
            yfs.asset_types.AssetTypes enum. If this error is raised please raise and issue on
            github with the output.
    """
    url = (
        "https://finance.yahoo.com/_finance_doubledown/api/"
        f"resource/searchassist;searchTerm={quote_lookup}"
    )

    response = requestor(url, **kwargs)

    if response.ok:

        try:
            search_response = ValidSymbolList(**response.json())

        except ValidationError as error:
            print("\n", error.json(indent=4), "\n")
            raise error

        else:
            if use_filter is True:
                valid_symbols = search_response.filter_symbols(exchange_type, asset_type)
            else:
                valid_symbols = search_response

            if first_ticker is True:
                return valid_symbols[0]  # returns single ValidSymbol

            return valid_symbols  # returns ValidSymbolList

    return None


#



"""Module for commonly used paths."""

from pathlib import Path

SOURCE_ROOT = Path(__file__).parent
"""* A path to the yfs source code directory."""

PROJECT_ROOT = SOURCE_ROOT.parent
"""* A path to the yfs project directory."""

TEST_DIRECTORY = PROJECT_ROOT / "tests"
"""* A path to the yfs test directory"""



"""Download multiple pages with or without threads."""

from concurrent.futures import as_completed, ThreadPoolExecutor
from typing import Callable, List, Optional

import enlighten
from pydantic import BaseModel as Base

from .lookup import fuzzy_search


def _download_pages_without_threads(  # pylint: disable=too-many-arguments
    group_object: Base,
    callable_: Callable,
    symbols: List[str],
    use_fuzzy_search: bool,
    page_not_found_ok: bool,
    progress_bar: bool,
    **kwargs,  # noqa: ANN003
) -> Optional[Base]:

    pages = group_object()

    if use_fuzzy_search:
        valid_symbols = []

        if progress_bar:
            pbar = enlighten.Counter(
                total=len(symbols), desc="Validating symbols...", unit="symbols"
            )

        for symbol in symbols:
            result = fuzzy_search(
                symbol,
                first_ticker=True,
                **kwargs,  # session, proxies, timeout
            )

            valid_symbols.append(result)

            if progress_bar:
                pbar.update()

        valid_symbols = filter(lambda s: s is not None, valid_symbols)
        symbols = list(set(s.symbol for s in valid_symbols))

    if progress_bar:
        pbar = enlighten.Counter(
            total=len(symbols), desc="Downloading Page Data...", unit="symbols"
        )

    for symbol in symbols:
        results = callable_(
            symbol,
            use_fuzzy_search=False,
            page_not_found_ok=page_not_found_ok,
            **kwargs,  # session, proxies, timeout
        )

        if results:
            pages.append(results)

        if progress_bar:
            pbar.update()

    if len(pages) > 0:
        return pages

    return None


def _download_pages_with_threads(  # pylint: disable=too-many-arguments
    group_object: Base,
    callable_: Callable,
    symbols: List[str],
    use_fuzzy_search: bool,
    page_not_found_ok: bool,
    thread_count: int,
    progress_bar: bool,
    **kwargs,  # noqa: ANN003
) -> Optional[Base]:
    pages = group_object()

    if use_fuzzy_search:
        valid_symbols = []

        with ThreadPoolExecutor(max_workers=thread_count) as executor:
            futures = [
                executor.submit(
                    fuzzy_search,
                    symbol,
                    first_ticker=True,
                    **kwargs,
                    # kwargs for requestor: session, proxies, timeout
                )
                for symbol in symbols
            ]

            if progress_bar:
                pbar = enlighten.Counter(
                    total=len(futures), desc="Validating symbols...", unit="symbols"
                )

            for future in as_completed(futures):
                valid_symbols.append(future.result())

                if progress_bar:
                    pbar.update()

        valid_symbols = filter(lambda s: s is not None, valid_symbols)
        symbols = list(set(s.symbol for s in valid_symbols))

    with ThreadPoolExecutor(max_workers=thread_count) as executor:
        futures = [
            executor.submit(
                callable_,
                symbol,
                use_fuzzy_search=False,
                page_not_found_ok=page_not_found_ok,
                **kwargs,
                # kwargs for requestor: session, proxies, timeout
            )
            for symbol in symbols
        ]

        if progress_bar:
            pbar = enlighten.Counter(
                total=len(futures), desc="Downloading Page Data...", unit="symbols"
            )

        for future in as_completed(futures):
            results = future.result()

            if results:
                pages.append(results)

            if progress_bar:
                pbar.update()

    if len(pages) > 0:
        return pages

    return None



"""Contains the classes and functions for scraping a yahoo finance statistics page."""

from collections import ChainMap
from enum import Enum
from typing import Iterable, List, Optional

import numpy as np
import pandas
from pandas import DataFrame
from pendulum.date import Date
from pydantic import BaseModel as Base
from pydantic import Field
from requests_html import HTML

from .cleaner import cleaner, CommonCleaners, field_cleaner, table_cleaner
from .lookup import fuzzy_search
from .multidownloader import _download_pages_with_threads, _download_pages_without_threads
from .quote import parse_quote_header_info, Quote
from .requestor import requestor


class PeriodType(str, Enum):
    """Enum which describes the period the data represents."""

    ANNUALLY = "Annually"
    QUARTERLY = "Quarterly"
    MONTHLY = "Monthly"


class Valuation(Base):
    """Data representing the intrinsic value of an asset for a single date.

    This is for one column of an entire valuation table.

    Attributes:
        date (Date): Date of the valuation.
        period_type (PeriodType): Annual, Quarterly, or Monthly.
        market_cap_intraday (int): Calculated using shares_outstanding from the
            most recently filed report.
        enterprise_value (int): Measure of a company's total value.
        trailing_pe (float): Relative valuation multiple that is based on the
            last 12 months of actual earnings
        forward_pe (float): Forward price-to-earnings (forward P/E) is a version of the
            ratio of price-to-earnings (P/E) that uses forecasted earnings for the P/E calculation.
        peg_ratio_five_year_expected (float): A valuation metric for determining
            the relative trade-off between the price of a stock, the EPS, and the
            company's expected growth.
        price_sales_ttm (float): Trailing Twelve Months price to sales ratio.
        price_book_mrq (float): todo
        enterprise_revenue (float): todo
        enterprise_ebitda (float): todo

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.
    """

    date: Date
    period_type: PeriodType

    market_cap_intraday: Optional[int] = Field(alias="market_cap_intraday_5")
    enterprise_value: Optional[int] = Field(alias="enterprise_value_3")
    trailing_pe: Optional[float] = Field(alias="trailing_p_e")
    forward_pe: Optional[float] = Field(alias="forward_p_e_1")
    peg_ratio_five_year_expected: Optional[float] = Field(alias="peg_ratio_five_year_expected_1")
    price_sales_ttm: Optional[float]
    price_book_mrq: Optional[float]
    enterprise_revenue: Optional[float] = Field(alias="enterprise_value_revenue_3")
    enterprise_ebitda: Optional[float] = Field(alias="enterprise_value_ebitda_6")

    _clean_dates = cleaner("date")(CommonCleaners.clean_date)

    _clean_common_values = cleaner(
        "market_cap_intraday",
        "enterprise_value",
        "trailing_pe",
        "forward_pe",
        "peg_ratio_five_year_expected",
        "price_sales_ttm",
        "price_book_mrq",
        "enterprise_revenue",
        "enterprise_ebitda",
    )(CommonCleaners.clean_common_values)

    class Config:
        """Pydantic config."""

        use_enum_values = True


class ValuationMeasuresTable(Base):
    """Representing the entire Valuation Measures Table on a yahoo finance Statistics Page.

    Attributes:
        valuations (List[Valuation]): A list of valuation objects.

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.

    """

    valuations: List[Valuation]

    @property
    def dataframe(self) -> DataFrame:
        """Return the Valuation Measures Table as a dataframe."""
        data = self.dict()
        dataframe = DataFrame.from_dict(data["valuations"])
        dataframe.set_index("date", inplace=True)
        dataframe.sort_index(inplace=True)
        return dataframe


class FinancialHighlights(Base):
    """Financial highlights section of a yahoo finance statistics page.

    Attributes:
        fiscal_year_ends (Date): todo
        most_recent_quarter_mrq (Date): todo

        profit_margin (float): todo
        operating_margin_ttm (float): todo

        return_on_assets_ttm (float): todo
        return_on_equity_ttm (float): todo

        revenue_ttm (int): todo
        revenue_per_share_ttm (float): todo
        quarterly_revenue_growth_yoy (float): todo
        gross_profit_ttm (int): todo
        ebitda (int): todo
        net_income_avi_to_common_ttm (int): todo
        diluted_eps_ttm (float): todo
        quarterly_earnings_growth_yoy (float): todo

        total_cash_mrq (int): todo
        total_cash_per_share_mrq (float): todo
        total_debt_mrq (int): todo
        total_debt_equity_mrq (float): todo
        current_ratio_mrq (float): todo
        book_value_per_share_mrq (float): todo

        levered_free_cash_flow_ttm (int): todo
        operating_cash_flow_ttm (int): todo

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.

    """

    fiscal_year_ends: Optional[Date]
    most_recent_quarter_mrq: Optional[Date]

    profit_margin: Optional[float]
    operating_margin_ttm: Optional[float]

    return_on_assets_ttm: Optional[float]
    return_on_equity_ttm: Optional[float]

    revenue_ttm: Optional[int]
    revenue_per_share_ttm: Optional[float]
    quarterly_revenue_growth_yoy: Optional[float]
    gross_profit_ttm: Optional[int]
    ebitda: Optional[int]
    net_income_avi_to_common_ttm: Optional[int]
    diluted_eps_ttm: Optional[float]
    quarterly_earnings_growth_yoy: Optional[float]

    total_cash_mrq: Optional[int]
    total_cash_per_share_mrq: Optional[float]
    total_debt_mrq: Optional[int]
    total_debt_equity_mrq: Optional[float]
    current_ratio_mrq: Optional[float]
    book_value_per_share_mrq: Optional[float]

    levered_free_cash_flow_ttm: Optional[int]
    operating_cash_flow_ttm: Optional[int]

    _clean_dates = cleaner("fiscal_year_ends", "most_recent_quarter_mrq")(
        CommonCleaners.clean_date
    )

    _clean_percentages = cleaner(
        "profit_margin",
        "operating_margin_ttm",
        "return_on_assets_ttm",
        "return_on_equity_ttm",
        "quarterly_revenue_growth_yoy",
        "quarterly_earnings_growth_yoy",
    )(CommonCleaners.clean_basic_percentage)

    _clean_common_values = cleaner(
        "book_value_per_share_mrq",
        "revenue_ttm",
        "revenue_per_share_ttm",
        "total_cash_mrq",
        "gross_profit_ttm",
        "ebitda",
        "net_income_avi_to_common_ttm",
        "diluted_eps_ttm",
        "total_cash_per_share_mrq",
        "total_debt_mrq",
        "total_debt_equity_mrq",
        "current_ratio_mrq",
        "levered_free_cash_flow_ttm",
        "operating_cash_flow_ttm",
    )(CommonCleaners.clean_common_values)


class TradingInformation(Base):
    """Trading information section of a yahoo finance statistics page.

    Attributes:
        beta_five_year_monthly (float): todo

        fifty_two_week_change (float): todo
        sp500_fifty_two_week_change (float): todo
        fifty_two_week_high (float): todo
        fifty_two_week_low (float): todo
        fifty_day_moving_average (float): todo
        two_hundred_day_moving_average (float): todo

        average_three_month_volume (int): todo
        average_ten_day_volume (int): todo
        shares_outstanding (int): todo
        float (int): todo
        percent_held_by_insiders (float): todo
        percent_held_by_institutions (float): todo

        shares_short (int): todo
        shares_short_date (Date): todo
        short_ratio (float): todo
        short_ratio_date (Date): todo
        short_percent_of_float (float): todo
        short_percent_of_float_date (Date): todo
        short_percent_of_shares_outstanding (float): todo
        short_percent_of_shares_outstanding_date (Date): todo
        shares_short_prior_month (int): todo
        shares_short_prior_month_date (Date): todo

        forward_annual_dividend_rate (float): todo
        forward_annual_dividend_yield (float): todo
        trailing_annual_dividend_rate (float): todo
        trailing_annual_dividend_yield (float): todo

        five_year_average_dividend_yield (float): todo

        payout_ratio (float): todo
        dividend_date (Date): todo
        exdividend_date (Date): todo
        last_split_factor (str): todo
        last_split_date (Date): todo

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.

    """

    beta_five_year_monthly: Optional[float]

    fifty_two_week_change: Optional[float] = Field(alias="fifty_twoweek_change_3")
    sp500_fifty_two_week_change: Optional[float] = Field(alias="sp500_fifty_twoweek_change_3")
    fifty_two_week_high: Optional[float] = Field(alias="fifty_two_week_high_3")
    fifty_two_week_low: Optional[float] = Field(alias="fifty_two_week_low_3")
    fifty_day_moving_average: Optional[float] = Field(alias="fifty_day_moving_average_3")
    two_hundred_day_moving_average: Optional[float] = Field(
        alias="two_hundred_day_moving_average_3"
    )

    average_three_month_volume: Optional[int] = Field(alias="avg_vol_three_month_3")
    average_ten_day_volume: Optional[int] = Field(alias="avg_vol_ten_day_3")
    shares_outstanding: Optional[int] = Field(alias="shares_outstanding_5")
    float: Optional[int]
    percent_held_by_insiders: Optional[float] = Field(alias="percent_held_by_insiders_1")
    percent_held_by_institutions: Optional[float] = Field(alias="percent_held_by_institutions_1")

    shares_short: Optional[int]
    shares_short_date: Optional[Date]
    short_ratio: Optional[float]
    short_ratio_date: Optional[Date]
    short_percent_of_float: Optional[float]
    short_percent_of_float_date: Optional[Date]
    short_percent_of_shares_outstanding: Optional[float]
    short_percent_of_shares_outstanding_date: Optional[Date]
    shares_short_prior_month: Optional[int]
    shares_short_prior_month_date: Optional[Date]

    forward_annual_dividend_rate: Optional[float] = Field(alias="forward_annual_dividend_rate_4")
    forward_annual_dividend_yield: Optional[float] = Field(alias="forward_annual_dividend_yield_4")
    trailing_annual_dividend_rate: Optional[float] = Field(alias="trailing_annual_dividend_rate_3")
    trailing_annual_dividend_yield: Optional[float] = Field(
        alias="trailing_annual_dividend_yield_3"
    )

    five_year_average_dividend_yield: Optional[float] = Field(
        alias="five_year_average_dividend_yield_4"
    )
    payout_ratio: Optional[float] = Field(alias="payout_ratio_4")
    dividend_date: Optional[Date] = Field(alias="dividend_date_3")
    exdividend_date: Optional[Date] = Field(alias="exdividend_date_4")
    last_split_factor: Optional[str] = Field(alias="last_split_factor_2")  # TODO: Not sure yet
    last_split_date: Optional[Date] = Field(alias="last_split_date_3")

    _clean_common_values = cleaner(
        "beta_five_year_monthly",
        "fifty_two_week_high",
        "fifty_two_week_low",
        "fifty_day_moving_average",
        "two_hundred_day_moving_average",
        "average_three_month_volume",
        "average_ten_day_volume",
        "shares_outstanding",
        "float",
        "shares_short",
        "short_ratio",
        "shares_short_prior_month",
        "forward_annual_dividend_rate",
        "trailing_annual_dividend_rate",
        "five_year_average_dividend_yield",
    )(CommonCleaners.clean_common_values)

    _clean_percentages = cleaner(
        "fifty_two_week_change",
        "sp500_fifty_two_week_change",
        "percent_held_by_insiders",
        "percent_held_by_institutions",
        "short_percent_of_float",
        "short_percent_of_shares_outstanding",
        "forward_annual_dividend_yield",
        "trailing_annual_dividend_yield",
        "payout_ratio",
    )(CommonCleaners.clean_basic_percentage)

    _clean_date = cleaner(
        "shares_short_date",
        "short_ratio_date",
        "short_percent_of_float_date",
        "short_percent_of_shares_outstanding_date",
        "shares_short_prior_month_date",
        "dividend_date",
        "exdividend_date",
        "last_split_date",
    )(CommonCleaners.clean_date)


def parse_valuation_table(
    html: HTML, period_type: PeriodType = PeriodType.QUARTERLY
) -> Optional[ValuationMeasuresTable]:
    """Parse and clean fields and rows of a valuation measures table HTML element.

    Args:
        html: Html element containing valuation table data.
        period_type (PeriodType): The period to be parsed. Only quarterly is currently supported.

    Returns:
        ValuationMeasuresTable: If data is found.
        None: No data available.
    """
    # IDEA: Parse the period type based on if it is a link or not.
    def clean_date(date_: str) -> str:
        """Clean field of a valuation table with date."""
        return date_.replace("Current", "").replace("As of Date:", "").strip()

    table_element = html.find(r"table.W\(100\%\).Bdcl\(c\).M\(0\).Whs\(n\).D\(itb\)", first=True)
    if table_element:

        table = pandas.read_html(table_element.html, index_col=0)
        table = table[0].transpose()
        table = table.replace(np.nan, "N/A", regex=True)

        valuations = []

        data = {}

        data["period_type"] = period_type
        for date_, row in table.iterrows():
            data["date"] = clean_date(date_)

            for field, value in row.items():
                field = field_cleaner(field)
                data[field] = value

            valuations.append(Valuation(**data))

        return ValuationMeasuresTable(valuations=valuations)

    return None


def parse_financial_highlights_table(html: HTML) -> Optional[FinancialHighlights]:
    """Parse and clean fields and rows of a financial highlights section of an HTML element."""
    table = html.find(r".Mb\(10px\).Pend\(20px\).smartphone_Pend\(0px\)", first=True)

    if table:
        table_data = table_cleaner(table)

        return FinancialHighlights(**table_data)

    return None


def parse_trading_information_table(html: HTML) -> Optional[TradingInformation]:
    """Parse and clean fields and rows of a trading information section of an HTML element."""
    table_element = html.find(r".Fl\(end\).W\(50\%\).smartphone_W\(100\%\)", first=True)

    if table_element:

        rows = [row.text.split("\n") for row in table_element.find("tr")]
        rows = list(filter(lambda row: len(row) == 2, rows))

        table_data = table_cleaner(html)

        if not table_data:
            table_data = {}

        for field, value in rows:
            if CommonCleaners.value_is_missing(value):
                continue

            if "Short" in field:
                field = field.replace("(prior month ", "prior month (")

                field_name, date_ = field.split("(")

                field_name = field_cleaner(field_name)
                date_ = date_.replace(") 4", "")

                table_data[field_name + "date"] = date_
                table_data[field_name.strip("_")] = value

        if table_data:
            return TradingInformation(**table_data)

    return None


class StatisticsPage(Base):
    """Represents all data you can find on a yahoo finance statistics page.

    Attributes:
        symbol (st): Ticker Symbol.
        quote (Quote): Quote data from the quote header section.
        valuation_measures (ValuationMeasuresTable): Valuation Measures Table section.
        financial_highlights (FinancialHighlights): Financial Highlights section.
        trading_information (TradingInformation): Trading Information section

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.

    """

    symbol: str
    quote: Quote
    valuation_measures: ValuationMeasuresTable
    financial_highlights: FinancialHighlights
    trading_information: TradingInformation

    def __lt__(self, other) -> bool:  # noqa: ANN001
        """Compare StatisticsPage objects to allow ordering by symbol."""
        if other.__class__ is self.__class__:
            return self.symbol < other.symbol

        return None


class StatisticsPageGroup(Base):
    """Multiple Statistics Pages Group together.

    Attributes:
        pages: Multiple StatisticsPage objects. A page for each symbol requested.

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.

    """

    pages: List[StatisticsPage] = list()

    def append(self, page: StatisticsPage) -> None:
        """Append a StatisticsPage to the StatisticsPageGroup.

        Args:
            page (StatisticsPage): A StatisticsPage object to add to the group.
        """
        if page.__class__ is StatisticsPage:
            self.pages.append(page)
        else:
            raise AttributeError("Can only append StatisticsPage objects.")

    @property
    def symbols(self: "StatisticsPageGroup") -> List[str]:
        """List of symbols in the StatisticsPageGroup."""
        return [symbol.symbol for symbol in self]

    def sort(self: "StatisticsPageGroup") -> None:
        """Sort StatisticsPage objects by symbol."""
        self.pages = sorted(self.pages)

    @property
    def dataframe(self: "StatisticsPageGroup") -> DataFrame:
        """Return a dataframe of multiple statistics pages."""
        data = []
        for page in self.pages:
            chain_map = ChainMap(
                page.quote.dict(),
                page.financial_highlights.dict(),
                page.trading_information.dict(),
            )
            chain_map["symbol"] = page.symbol
            data.append(chain_map)

        dataframe = DataFrame.from_dict(data)
        dataframe.set_index("symbol", inplace=True)
        dataframe.sort_index(inplace=True)
        return dataframe

    def __iter__(self: "StatisticsPage") -> Iterable:
        """Iterate over StatisticsPage objects."""
        return iter(self.pages)

    def __len__(self: "StatisticsPageGroup") -> int:
        """Length of StatisticsPage objects."""
        return len(self.pages)


def get_statistics_page(
    symbol: str,
    use_fuzzy_search: bool = True,
    page_not_found_ok: bool = False,
    **kwargs,  # noqa: ANN003
) -> Optional[StatisticsPage]:
    """Get statistics page data.

    Args:
        symbol (str): Ticker symbol.
        use_fuzzy_search (bool): If True validates symbol prior to requesting options page data.
        page_not_found_ok (bool): If True Returns None when page is not found.
        **kwargs: Pass (session, proxies, and timeout) to the requestor function.

    Returns:
        StatisticsPage: When data is found.
        None: No data is found and page_not_found_ok is True.

    Raises:
        AttributeError: When a page is not found and the page_not_found_ok arg is false.
    """
    if use_fuzzy_search:
        fuzzy_response = fuzzy_search(symbol, first_ticker=True, **kwargs)

        if fuzzy_response:
            symbol = fuzzy_response.symbol

    url = f"https://finance.yahoo.com/quote/{symbol}/key-statistics?p={symbol}"

    response = requestor(url)

    if response.ok:

        html = HTML(html=response.text)

        quote = parse_quote_header_info(html)
        valulation_measures = parse_valuation_table(html)
        financial_highlights = parse_financial_highlights_table(html)
        trading_information = parse_trading_information_table(html)

        if quote and valulation_measures and financial_highlights and trading_information:

            return StatisticsPage(
                symbol=symbol,
                quote=quote,
                valuation_measures=valulation_measures,
                financial_highlights=financial_highlights,
                trading_information=trading_information,
            )
    if page_not_found_ok:
        return None

    raise AttributeError(f"{symbol} statistics page not found.")


def get_multiple_statistics_pages(  # pylint: disable=too-many-arguments
    symbols: List[str],
    use_fuzzy_search: bool = True,
    page_not_found_ok: bool = True,
    with_threads: bool = False,
    thread_count: int = 5,
    progress_bar: bool = True,
    **kwargs,  # noqa: ANN003
) -> Optional[StatisticsPageGroup]:
    """Get multiple statistics pages.

    Args:
        symbols (List[str]): Ticker symbols or company names.
        use_fuzzy_search (bool): If True does a symbol lookup validation prior
            to requesting data.
        page_not_found_ok (bool): If True Returns None when page is not found.
        with_threads (bool): If True uses threading.
        thread_count (int): Number of threads to use if with_threads is set to True.
        **kwargs: Pass (session, proxies, and timeout) to the requestor function.
        progress_bar (bool): If True shows the progress bar else the progress bar
            is not shown.

    Returns:
        StatisticsPageGroup: When data is found.
        None: No data is found and page_not_found_ok is True.

    Raises:
        AttributeError: When a page is not found and the page_not_found_ok arg is false.
    """
    symbols = list(set(symbols))
    group_object = StatisticsPageGroup
    callable_ = get_statistics_page

    if with_threads:
        return _download_pages_with_threads(
            group_object,
            callable_,
            symbols,
            use_fuzzy_search=use_fuzzy_search,
            page_not_found_ok=page_not_found_ok,
            thread_count=thread_count,
            progress_bar=progress_bar,
            **kwargs,
        )
    return _download_pages_without_threads(
        group_object,
        callable_,
        symbols,
        use_fuzzy_search=use_fuzzy_search,
        page_not_found_ok=page_not_found_ok,
        progress_bar=progress_bar,
        **kwargs,
    )



"""Module for parsing quote header data from a yahoo finance page."""

from typing import Optional

from pydantic import BaseModel as Base
from requests_html import HTML

from .cleaner import cleaner, CommonCleaners


def clean_quote_name(value: str) -> str:
    """Remove the symbol and strip whitespace from the company name.

    Args:
        value (str): A company name string parsed from the quote header
            portion of any yahoo finance page.

    Example:
        |Input             |Output    |
        |------------------|----------|
        |Apple Inc. (AAPL) |Apple Inc.|

    Returns:
        str: Company name with the ticker symbol removed.

    """
    value = value.split("(")
    return value[0].strip()


class Quote(Base):
    """Represents all the data parsed from the quote header section of a yahoo finance page.

    Attributes:
        name (str): Company name.
        close (float): Close price.
        change (float): Dollar change in price.
        percent_change (float): Percent change in price.

    Notes:
        This class inherits from the pydantic BaseModel which allows for the use
        of .json() and .dict() for serialization to json strings and dictionaries.

        .json(): Serialize to a JSON object.
        .dict(): Serialize to a dictionary.
    """

    name: str
    close: Optional[float]
    change: Optional[float]
    percent_change: Optional[float]

    _clean_name = cleaner("name")(clean_quote_name)

    _clean_close = cleaner("close")(CommonCleaners.clean_common_values)

    _clean_change = cleaner("change")(CommonCleaners.clean_first_value_split_by_space)

    _clean_percent_change = cleaner("percent_change")(
        CommonCleaners.clean_second_value_split_by_space
    )


def parse_quote_header_info(html: HTML) -> Optional[Quote]:
    """Parse and clean html elements from the quote header info portion of a yahoo finance page.

    Args:
        html (HTML): An HTML object containing quote header info data ready to be parse.

    Returns:
        Quote: Quote object containing the parsed quote header data if successfully parsed.
        None: No quote header info data present in the HTML.
    """
    quote_selectors = {
        "name": r".D\(ib\).Fz\(18px\)",
        "close": r".Trsdu\(0\.3s\).Fw\(b\).Fz\(36px\).Mb\(-4px\).D\(ib\)",
        "change": r".Trsdu\(0\.3s\).Fw\(500\)",
        "percent_change": r".Trsdu\(0\.3s\).Fw\(500\)",
    }

    quote_header_info = html.find("div#quote-header-info", first=True)

    quote_data = {}

    if quote_header_info:

        for field, selector in quote_selectors.items():
            element = quote_header_info.find(selector)

            if element and len(element) == 1:
                quote_data[field] = element[0].text

    if quote_data:
        return Quote(**quote_data)

    return None



"""Module for exchanges organized by region."""

from dataclasses import dataclass
from enum import Enum
from typing import Union

from more_itertools import flatten


class UnitedStatesExchanges(str, Enum):
    """Exchanges based out of the United States."""

    CBOE = "Chicago Board Options Exchange"
    CBOT = "Chicago Board of Trade"
    CME = "Chicago Mercantile Exchange"
    DOW = "Dow Jones"
    NASDAQ = "NASDAQ"
    NASDAQ_GIDS = "NASDAQ GIDS"
    NSE = "NSE"
    NY_MERCANTILE = "NY Mercantile"
    NYBOT = "New York Board of Trade"
    NYCE = "New York Commodity Exchange"
    NYSE = "NYSE"
    NYSE_MKT = "NYSE MKT"
    NYSE_ARCA = "NYSEArca"
    OTC = "OTC Markets"
    OTC_BB = "OTC BB"


class CanadianExchanges(str, Enum):
    """Exchanges Based out of Canda."""

    NATURAL_RESOURCES = "CNQ"
    NEO = "NEO"
    TORONTO = "Toronto"
    VENTURE = "CDNX"


class SouthAmericanExchanges(str, Enum):
    """Exchanges Based out of South America."""

    BUENOS_AIRES = "Buenos Aires"
    MEXICO = "Mexico"
    SAO_PAOLO = "Sao Paolo"
    SANTIAGO = "Santiago Stock Exchange"


class EuropeanExchanges(str, Enum):
    """Exchanges Based out of Europe."""

    AMSTERDAM = "Amsterdam"
    ATHENS = "Athens"
    BERLIN = "Berlin"
    BRUSSELS = "Brussels Stock Exchange"
    COPENHAGEN = "Copenhagen"
    DUSSELDORF = "Dusseldorf Stock Exchange"
    EURONEXT = "Euronext"
    FTSE_GLOBAL_INDEX = "FTSEGlobal Index"
    FRANKFURT = "Frankfurt"
    HAMBURG = "Hamburg"
    HANOVER = "Hanover"
    HELSINKI = "HEL"
    IRISH = "Irish"
    ISTANBOL = "IST"
    LISBON = "Lisbon Stock Exchange"
    LONDON = "London"
    LONDON_INTERNATIONAL = "International Orderbook - London"
    MADRID = "Madrid Stock Exchange CATS"
    MILAN = "Milan"
    MUNICH = "Munich"
    NORWAY = "Oslo"
    PARIS = "Paris"
    PRAGUE = "Prague Stock Exchange"
    STOCKHOLM = "Stockholm"
    STUTTGART = "Stuttgart"
    SWISS = "Swiss"
    VIENNA = "Vienna"
    ZURICH = "Zurich Stock Exchange"


class AfricanExchanges(str, Enum):
    """Exchanges Based out of Africa."""

    CAIRO = "Cairo Stock Exchange"
    JOHANNESBURG = "Johannesburg Stock Exchange"


class MiddleEasternExchanges(str, Enum):
    """Exchanges Based out of MiddleEast."""

    PAKISTAN = "PSX"
    SAUDI = "Saudi Stock Exchange"
    TEL_AVIV = "Tel Aviv"


class AsianExchanges(str, Enum):
    """Exchanges Based out of Asia."""

    BOMBAY = "Bombay"
    COLOMBO = "Colombo Stock Exchange"
    HONG_KONG = "Hong Kong"
    JAKARTA = "Jakarta"
    KOREA = "Korea"
    KOSDAQ = "KOSDAQ"
    KUALA_LUMPUR = "Kuala Lumpur Stock Exchange"
    SHENZHEN = "Shenzhen"
    SHANGHAI = "Shanghai"
    SINGAPORE = "Singapore"
    TAIWAN = "Taiwan"
    THAILAND = "SET"
    TOKYO = "Tokyo Stock Exchange"


class AustralianExchanges(str, Enum):
    """Exchanges Based out of Australia."""

    AUSTRALIAN = "Australian"
    NEW_ZEALAND = "New Zealand"


class UnknownExchanges(str, Enum):
    """Exchanges that haven't been identified yet."""

    BUD = "BUD"
    SNP = "SNP"
    OPR = "OPR"
    CCC = "CCC"
    CCY = "CCY"
    MULTI_COMMODITY = "MCX"
    TAL = "TAL"
    LIT = "LIT"
    RIS = "RIS"
    TLX = "TLX Exchange"
    INDUSTRY = "Industry"
    XETRA = "XETRA"
    UNKNOWN = "UNKNOWN"  # placeholder in case all exchanges are identified.


VALID_EXCHANGE_UNION = Union[  # pylint: disable=invalid-name
    UnitedStatesExchanges,
    CanadianExchanges,
    AustralianExchanges,
    AsianExchanges,
    SouthAmericanExchanges,
    EuropeanExchanges,
    MiddleEasternExchanges,
    AfricanExchanges,
    UnknownExchanges,
]

VALID_EXCHANGE_ENUMS = list(
    flatten(
        [
            list(UnitedStatesExchanges),
            list(CanadianExchanges),
            list(AustralianExchanges),
            list(AsianExchanges),
            list(SouthAmericanExchanges),
            list(EuropeanExchanges),
            list(MiddleEasternExchanges),
            list(AfricanExchanges),
            list(UnknownExchanges),
        ]
    )
)

VALID_EXCHANGE_ENUM_NAMES = [exchange.name for exchange in VALID_EXCHANGE_ENUMS]

VALID_EXCHANGE_ENUM_VALUES = [exchange.value for exchange in VALID_EXCHANGE_ENUMS]


@dataclass
class ExchangeTypes:  # pylint: disable=too-many-instance-attributes, no-member
    """Helper for choosing a valid exchange region.

    This is used for filtering out results from a fuzzy_search request.

    Attributes:
        united_states (Enum): UnitedStatesExchanges
        canada (Enum): CanadianExchanges
        australian (Enum): AustralianExchanges
        asia (Enum): AsianExchanges
        south_america (Enum): SouthAmericanExchanges
        europe (Enum): EuropeanExchanges
        middle_east (Enum): MiddleEasternExchanges
        africa (Enum): AfricanExchanges
        unknown (Enum): UnknownExchanges

    """

    united_states: Enum = UnitedStatesExchanges
    canada: Enum = CanadianExchanges
    australian: Enum = AustralianExchanges
    asia: Enum = AsianExchanges
    south_america: Enum = SouthAmericanExchanges
    europe: Enum = EuropeanExchanges
    middle_east: Enum = MiddleEasternExchanges
    africa: Enum = AfricanExchanges
    unknown: Enum = UnknownExchanges

    @classmethod
    def show(cls) -> None:
        """Print out all valid exchanges."""
        exchange_types = cls().__dataclass_fields__  #
        print("Exchange Types:\n")

        for exchanges in exchange_types:
            print(exchanges)
        print()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['yfs']

package_data = \
{'': ['*']}

install_requires = \
['enlighten>=1.6.2,<2.0.0',
 'loguru>=0.5.2,<0.6.0',
 'more_itertools>=8.5.0,<9.0.0',
 'pandas>=1.1.2,<2.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'prompt-toolkit>=3.0.7,<4.0.0',
 'pydantic>=1.6.1,<2.0.0',
 'python-decouple>=3.3,<4.0',
 'requests-html>=0.10.0,<0.11.0',
 'requests[socks]>=2.24.0,<3.0.0']

setup_kwargs = {
    'name': 'yfs',
    'version': '0.3.0',
    'description': 'Scrape Yahoo Finance',
    'long_description': '# YFS\n## Yahoo Finance Scraper with a WYSIWYG approach.\n[![Travis (.org)](https://img.shields.io/travis/dgnsrekt/yfs?style=for-the-badge&logo=appveyor?url=)](https://travis-ci.com/dgnsrekt/yfs)\n[![Black](https://img.shields.io/badge/code%20style-black-black?style=for-the-badge&logo=appveyor)](https://github.com/psf/black)\n[![GitHub](https://img.shields.io/github/license/dgnsrekt/yfs?style=for-the-badge)](https://raw.githubusercontent.com/dgnsrekt/yfs/master/LICENSE)\n\nWant to scrape data from the summary page use the get_summary_page function. Want to scrape the summary pages of a list of symbols use the get_multiple_summary_pages function. The returned object can be serialized with .json, .dict, ***and depending on the object .dataframe*** methods. Each function can accept a proxy to help avoid rate limiting. In fact in the future you can install [requests-whaor](https://github.com/dgnsrekt/requests-whaor) ***ANOTHER WORK IN PROGRESS*** which supplies a rotating proxy to bypass rate limits.\n\n***Before you start please note adding historical data to this API is not a priority. At some point I will get around to it. My main focus are options, quote information and symbol lookup. So please do not raise issues about historical data.***\n\n## Features\n* [x] Company and Symbol lookup\n* [x] Summary Page\n* [x] Option Chains\n* [x] Statistics Page\n\n## [>> To the documentation.](https://dgnsrekt.github.io/yfs/)\n\n## Quick Start\n\n### Prereqs\n* Python ^3.8\n\n### Install with pip\n```\npip install yfs\n```\n\n### Install with [poetry](https://python-poetry.org/)\n```\npoetry add yfs\n```\n\n### How to scrape multiple summary pages from yahoo finance.\n```python\nfrom yfs import get_multiple_summary_pages\n\nsearch_items = ["Apple", "tsla", "Microsoft", "AMZN"]\n\nsummary_results = get_multiple_summary_pages(search_items)\nfor page in summary_results:\n    print(page.json(indent=4))\n    break  # To shorten up the quick-start output.\n\nCOLUMNS = [\n    "close",\n    "change",\n    "percent_change",\n    "average_volume",\n    "market_cap",\n    "earnings_date",\n]\nprint(summary_results.dataframe[COLUMNS])\n\n```\nOutput\n```python\n➜ python3 quick_start_example.py\nDownloading Summary Data... 100%|██████████████████████████████████████████████████████████████████████████████████████████████████████████████████████████| 4/4 [00:03<00:00, 1.19 symbols/s]{\n    "symbol": "AAPL",\n    "name": "Apple Inc.",\n    "quote": {\n        "name": "Apple Inc.",\n        "close": 113.02,\n        "change": -3.77,\n        "percent_change": -3.23\n    },\n    "open": 112.89,\n    "high": 112.22,\n    "low": 115.37,\n    "close": 113.02,\n    "change": -3.77,\n    "percent_change": -3.23,\n    "previous_close": 116.79,\n    "bid_price": 112.58,\n    "bid_size": 800,\n    "ask_price": 112.77,\n    "ask_size": 3000,\n    "fifty_two_week_low": 137.98,\n    "fifty_two_week_high": 53.15,\n    "volume": 144711986,\n    "average_volume": 172065562,\n    "market_cap": 1933000000000,\n    "beta_five_year_monthly": 1.28,\n    "pe_ratio_ttm": 34.29,\n    "eps_ttm": 3.3,\n    "earnings_date": "2020-10-28",\n    "forward_dividend_yield": 0.82,\n    "forward_dividend_yield_percentage": 0.7,\n    "exdividend_date": "2020-08-07",\n    "one_year_target_est": 119.24\n}\n          close  change  percent_change  average_volume     market_cap earnings_date\nsymbol\nAAPL     113.02   -3.77           -3.23       172065562  1933000000000    2020-10-28\nAMZN    3125.00  -96.26           -2.99         5071692  1565000000000    2020-10-29\nMSFT     206.19   -6.27           -2.95        34844893  1560000000000    2020-10-21\nTSLA     415.09  -33.07           -7.38        80574089   386785000000    2020-10-21\n```\n### Next step [fuzzy search examples](https://dgnsrekt.github.io/yfs/examples/fuzzy-search-examples/)\n\n## TODO\n* [ ] More testing\n* [ ] More Docs\n* [ ] More examples\n* [ ] [WHAOR](https://github.com/dgnsrekt/requests-whaor) Example\n* [ ] Profile Page\n* [ ] Financials Page\n* [ ] Analysis Page\n* [ ] Holders page\n* [ ] Sustainability Page\n* [ ] Historical Page\n* [ ] Chart Page\n* [ ] Conversations Page maybe ¯\\_(ツ)_/¯\n\n## Contact Information\nTelegram = Twitter = Tradingview = Discord = @dgnsrekt\n\nEmail = dgnsrekt@pm.me\n',
    'author': 'dgnsrekt',
    'author_email': 'dgnsrekt@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://dgnsrekt.github.io/yfs/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
